;;; rx-to-peg.el --- Transform simple rx forms into really equivalent pegs -*- lexical-binding: t -*-

;; Copyright (C) 2021 Free Software Foundation, Inc

;; For a peg implementation for Emacs see peg.el in Gnu Elpa.

;; The idea for the algorithm is from here:
;;   http://www.lua.inf.puc-rio.br/publications/medeiros11regular.pdf
;;
;; We use the terms "peg" and "grammar" synonymously.  A peg is
;; represented as a list (r p) where r the list of rule definitions
;; and p the expression to match.
;;
;; Use like this:
;;
;; (rx-to-peg '(and (* (and (or "a" "aa") "b")) eos))
;;   ==>
;; (((a (or (and "ab"  a)
;;          (and "aab" a)
;;          (eob))))
;;  a)

;;; Code:

(require 'rx)

;; Helpers for handling grammars.

(defun rxtp--listify-for (op)
  (lambda (xs)
    (cl-mapcan (lambda (x) (if (eq (car-safe x) op) (cdr x) (list x)))
               xs)))

;; `rxtp--and' and `rxtp--or' help to build nice `and' and `or' expressions

(defun rxtp--and (&rest xs)
  (cl-flet* ((ensure-string (obj)
               (pcase obj
                 (`(char ,c) (string c))
                 (_ obj)))
             (stringishp (obj) (or (stringp obj) (eq 'char (car-safe obj))))
             (concat-subsequent-strings-in (list)
               (let ((rest list))
                 (while rest
                   (when (stringishp (car rest))
                     (while (stringishp (cadr rest))
                       (setcar rest (concat (ensure-string (car rest))
                                            (ensure-string (cadr rest))))
                       (setcdr rest (cddr rest))))
                   (cl-callf cdr rest)))
               list))
    (let ((operands (concat-subsequent-strings-in
                     (delete "" (funcall (rxtp--listify-for 'and) xs)))))
      (if (cdr operands)
          `(and ,@operands)
        (or (car operands) "")))))

(defun rxtp--or (&rest xs)
  (let ((operands (funcall (rxtp--listify-for 'or) xs)))
    (if (cdr operands)
        `(or ,@operands)
      (car operands))))

(defun rxtp--do (expr)
  ;; An immediately performed action for side effects only.  Never fails
  `(guard (progn ,expr t)))


(defvar rxtp-symbol-ctr)

(defun g--new-rule-name ()
  "Return a fresh non-terminal symbol."
  (let ((n (cl-incf rxtp-symbol-ctr)))
    (intern (if (<= n 26) (string (+ (- ?a 1) n)) (format "r%d" n)))))

(defvar rxtp--temp-rule-names)

(defun g--new-temp-rule-name ()
  (let ((d (cl-gensym "rule-")))
    (push d rxtp--temp-rule-names)
    d))

(defun rxtp-make-grammar (r p)
  (list r p))

(defun rxtp-grammar-rules (g)
  ;; Return G's rule defintions
  (car g))
(gv-define-simple-setter rxtp-grammar-rules setcar)

(defun rxtp-grammar-expression (g)
  ;; Return G's match expression
  (cadr g))

(defun rxtp-w/extended-expr (g f)
  ;; apply F to G's match expression and return the result as new peg
  (pcase-let ((`(,r ,p) g))
    `(,r ,(funcall f p))))

(defvar rxtp--group-associations)
(defvar rxtp--group-associations-used?)

(defun rxtp-w/renamed-rule (g oname nname)
  "Return a copy of G with rule ONAME renamed to NNAME."
  (cl-labels ((subst-rule (thing)
                (pcase-exhaustive thing
                  ((pred (eq oname)) nname)
                  ((or (pred symbolp) (pred vectorp) (pred stringp) (pred characterp) `(guard . ,_))
                   thing)
                  ((pred listp) (mapcar #'subst-rule thing)))))
    (subst-rule g)))


;; Definition of the transcription function phi

(defun rxtp-null ()
  "Return a trivial grammar with language {\"\"}."
  (rxtp-make-grammar () ""))

(defun rxtp--trivial-rx-p (rx)
  ;; All the stuff that can be translated directly goes here
  (pcase rx
    ((pred stringp)            rx)
    ((pred characterp)       `(char ,rx))
    ((or 'anychar 'anything) '(any))

    ;; beginning-of-*, end-of-* stuff
    ((or 'eos 'string-end   'eot 'buffer-end)   '(eob)) ;pegs search buffers, so eos gets eob
    ((or 'bos 'string-start 'bot 'buffer-start) '(bob)) ;likewise
    ((or 'bol 'line-start) '(bol))
    ((or 'eol 'line-end)   '(eol))
    ((or 'bow 'word-start) '(bow))
    ((or 'eow 'word-end)   '(eow))
    ('symbol-start         '(bos))
    ('symbol-end           '(eos))
    ('word-boundary        '(guard (looking-at "\\b")))
    ('not-word-boundary    '(guard (looking-at "\\B")))
    ('not-wordchar         '(and (guard (looking-at "\\W")) (any)))
    (`(syntax ,name)       `(syntax-class ,name))
    ((and (pred symbolp)
          (let (and (pred identity) c)
            (cdr (assq rx rx--char-classes))))
     (vector c))
    ((pred rx--charset-p)
     ;; good enough for now (real support would be some work...)
     (let* ((regexp-stringish (let ((print-escape-newlines t))
                                (prin1-to-string (rx-to-string rx))))
            (try-as-vector (and-let* (((string-match (rx bos "\"[" (group (* anychar)) "]\"" eos)
                                                     regexp-stringish))
                                      (ms (match-string 1 regexp-stringish)))
                             (if (string-match-p (rx bos "^" ) ms)
                                 `(and (not ,(vector (substring ms 1))) (any))
                               (vector ms)))))
       (or try-as-vector
           `(and (guard (looking-at ,regexp-stringish)) '(any)))))))

(defun rxtp--operator-greedyness (op)
  (cond
   ((memq op (list '*  '+  '\?  ?\s)) t)
   ((memq op (list '*? '+? '\?? ??))  nil)
   (t rx--greedy)))

(defun rxtp-phi (expr cont)
  ;; EXPR is a regular expression (an `rx' s-exp).  The second argument
  ;; CONT is a continuation grammar (when the function is recursively
  ;; called this argument is bound to the already transcribed
  ;; following parts of the grammar).
  ;;
  ;; The rewrite is so that when the following two conditions hold:
  ;;
  ;;   (1) For no sub-rx (* E) of EXPR does E match the empty string.
  ;;
  ;;   (2) The language of EXPR has the "prefix property": There are no
  ;;   distinct string S1 and S2 in the language of EXPR so that S1 is a
  ;;   prefix of S2.
  ;;
  ;; then the returned peg is equivalent to EXPR in the sense that both match the
  ;; same set of strings.  (2) can be ensured by matching the end of the
  ;; string (`eos' in `rx').

  (cl-flet ((cont-w/prepended (&rest exps)
              ;; prepend the peg EXPS to the expression of CONT
              (rxtp-w/extended-expr
               cont
               (lambda (p)
                 (apply #'rxtp--and (append exps (list p)))))))
    (pcase expr
      ((let (and (pred identity) transformed) (rxtp--trivial-rx-p expr))
       (cont-w/prepended transformed))

      ((or 'nonl 'not-newline 'any) (cont-w/prepended '(not ["\n"])) '(any))

      ;; `and' and `or'
      (`(,(or 'and 'seq 'sequence ':) . ,operands)
       (cl-reduce #'rxtp-phi operands :from-end t :initial-value cont))
      (`(,(or 'or '|) ,e1 . ,es)
       ;; Build a grammar that matches when one of (rxtp-phi E CONT) would
       ;; match - E is any expression in the given `or` - starting with E1
       (let ((g (rxtp-phi e1 cont)))
         (dolist (e2 es)
           (setq g (rxtp-w/extended-expr
                    (rxtp-phi
                     e2
                     (rxtp-make-grammar
                      (rxtp-grammar-rules g)
                      (rxtp-grammar-expression cont)))
                    (lambda (p2) (rxtp--or (rxtp-grammar-expression g)
                                           p2)))))
         g))

      ;; repetition operators (definition of `rxtp--**' see below)
      (`(,(and (or '* '*? (or 'zero-or-more '0+)) op) ,e)
       (rxtp--** e cont 0 'infinity (not (rxtp--operator-greedyness op))))
      (`(,(and (or '+ '+? (or 'one-or-more  '1+)) op) ,e)
       (rxtp--** e cont 1 'infinity (not (rxtp--operator-greedyness op))))
      (`(,(and (or '\? ?\s '\?? ?? 'zero-or-one 'opt 'optional) op) ,e)
       (rxtp--** e cont 0 1 (not (rxtp--operator-greedyness op))))
      (`(= ,n ,e)  (rxtp--** e cont n n))
      (`(>= ,n ,e) (rxtp--** e cont n 'infinity))
      (`(,(or '** 'repeat) ,n ,m ,e) (rxtp--** e cont n m))

      ;; Groups and Backrefs
      (`(,(or 'group 'submatch) . ,elts) (rxtp-phi `(and . ,elts) cont)) ;incomplete! (no ref created)
      (`(,(or 'group-n 'submatch-n) ,n . ,elts)
       (let ((g1 (rxtp-phi `(and . ,elts) (rxtp-null))))
         (rxtp-make-grammar
          (cl-union (rxtp-grammar-rules g1) (rxtp-grammar-rules cont))
          (rxtp--and
           ;; we save the ref in a gensym owned by the peg
           (rxtp--do `(setf (alist-get ,n ,rxtp--group-associations) (point)))
           (rxtp-grammar-expression g1)
           (rxtp--do
            `(cl-callf (lambda (start) (buffer-substring-no-properties start (point)))
                 (alist-get ,n ,rxtp--group-associations)))
                     (rxtp-grammar-expression cont)))))
      (`(backref ,n)
       (setq rxtp--group-associations-used? t)
       (rxtp-w/extended-expr
        cont
        (let ((s (make-symbol "s")))
          (lambda (e)
            (rxtp--and `(guard (let ((,s (alist-get ,n ,rxtp--group-associations)))
                                 (when (looking-at (regexp-quote ,s))
                                   (forward-char (length ,s))
                                   t)))
                       e)))))

      (unknown (error "Don't know how to handle `%S'" unknown)))))

(defun rxtp--** (e cont n m &optional non-greedy)
  ;; Transform repetitions of E, N to M times (M can be the symbol
  ;; `infinity`)
  ;; For a finite number of repetitions of a nontrivial expression we
  ;; use a separate rule matching it once.
  (let* ((loop* (and (eq m 'infinity) (g--new-temp-rule-name)))
         ;; a grammar matching E exactly once, using either a "leaf term" or
         ;; a single rule name as expression:
         (ge (if-let ((trivial-e-translation (rxtp--trivial-rx-p e)))
                 (rxtp-make-grammar () trivial-e-translation)
               (let ((g (rxtp-phi e (rxtp-null))))
                 (if (not (listp (rxtp-grammar-expression g))) g
                   (let ((rule-e (g--new-temp-rule-name)))
                     (rxtp-make-grammar (cons (list rule-e (rxtp-grammar-expression g))
                                              (rxtp-grammar-rules g))
                                        rule-e)))))))
    (cl-flet ((g--or (reverse operands)
                (apply #'rxtp--or (funcall (if reverse #'reverse #'identity) operands))))
      (rxtp-make-grammar
       (cl-reduce
        #'cl-union
        (list
         (and loop*
              (let ((loop-e-then-cont
                     (rxtp-phi e (rxtp-w/extended-expr cont (lambda (_) loop*)))))
                (list (list loop*
                            (g--or non-greedy
                                   (list (rxtp-grammar-expression loop-e-then-cont)
                                         (rxtp-grammar-expression cont)))))))
         (and (not (and (= n 0) (memq m (list 0 'infinity))))
              (rxtp-grammar-rules ge))
         (rxtp-grammar-rules cont)))
       (g--or (not non-greedy)
              (mapcar (lambda (k)
                        (apply #'rxtp--and
                               (append (make-list k (rxtp-grammar-expression ge))
                                       (list (or loop* (rxtp-grammar-expression cont))))))
                      (apply #'number-sequence (if loop* (list n) (list n m)))))))))

(declare-function xr "xr")

;; Our entry point:
(defun rx-to-peg (regexp)
  "Transform regular expression REGEXP into a PEG.
REGEXP is either a regular expression in string form or an
`rx' expression."
  (when (stringp regexp) (cl-callf xr regexp))
  (let ((rxtp--temp-rule-names ()) (peg) )
    (let ((rxtp--group-associations (make-symbol "group-associations"))
          (rxtp--group-associations-used? nil))
      (set rxtp--group-associations ())
      (setq peg
            ;; call phi with REGEXP and continuation grammar with language {""}
            (rxtp-phi (if (eq (car-safe regexp) 'rx) `(and ,@(cdr regexp)) regexp)
                      (rxtp-null)))
      (when rxtp--group-associations-used?
        (cl-callf rxtp-w/extended-expr peg
          (lambda (e) (rxtp--and `(guard (progn (setq ,rxtp--group-associations ()) t)) e))))
      ;; Try to make the result a bit less nasty:
      (cl-callf g--delete-equal-rules peg)
      (let ((rxtp-symbol-ctr 0))
        (dolist (rulename (mapcar #'car (rxtp-grammar-rules peg)))
          (cl-callf rxtp-w/renamed-rule peg rulename (g--new-rule-name)))))
    peg))

;;;; trivial beautifications

(defun g--rules-equiv-p (r1 r2)
  (equal (cdr r1) (cdr r2)))

(defun g--delete-equal-rules (g)
  (cl-maplist (pcase-lambda (`(,r1 . ,more-rules))
                (mapc (lambda (r2)
                        (when (g--rules-equiv-p r1 r2)
                          (cl-callf2 delq r2 (rxtp-grammar-rules g))
                          (cl-callf rxtp-w/renamed-rule g (car r2) (car r1))))
                      more-rules))
              (car g))
  g)



(provide 'rx-to-peg)
;;; rx-to-peg.el ends here
